package pqrtree;

/**
 * A <code>ColoredDoublyLinkedList</code>
 * is an object very similar to a <code>DoublyLinkedList</code>,
 * except by the kind of data its nodes stores, just
 * list nodes.
 * 
 * @version 0.1
 * @author Bruno Dilly
 * @see pqrtree.DoublyLinkedList
 */
public class ColoredDoublyLinkedList extends DoublyLinkedList {

	/**
	 * Instanciates a new colored children list.
	 * It's constructed just like <code>DoublyLinkedList</code>
	 * @see DoublyLinkedList#DoublyLinkedList()
	 */
	public ColoredDoublyLinkedList()
	{
		super();
	}

	/**
	 * Extracts colored children from the children list
	 * and creates a new list with them all.
	 * The children referece the new ancestor. 
	 * @param newAncestor The new ancestor of the
	 * extracted children.
	 * @param childList The child list from where the children
	 * will be removed.
	 * @return A new list with all the nodes extracted.
	 */
	public ChildrenDoublyLinkedList extractColoredChildren(PQRNode newAncestor, ChildrenDoublyLinkedList childList)
	{
		ChildrenDoublyLinkedList children = new ChildrenDoublyLinkedList();
		ListNode aux = head;
		ListNode iterator = startIteration();
		while (aux != null)
		{	
			ListNode auxchild = (ListNode)aux.getData();
			childList.removeNode(auxchild);
			((Node)auxchild.getData()).setParent(newAncestor);
			children.addNodeLast(auxchild);
			aux = aux.getNext(iterator);
		}
		return children;
	}
	
	/**
	 * Extracts colored children, except one, from the children list
	 * and creates a new list with them all.
	 * Removes the excluded list node from the colored children list.
	 * The children referece the new ancestor. 
	 * @param newAncestor The new ancestor of the
	 * extracted children.
	 * @param childList The child list from where the children
	 * will be removed.
	 * @param excluded The list node that will not be excluded, but will
	 * be removed from the colored children list.
	 * @return A new list with all the nodes extracted.
	 */
	public ChildrenDoublyLinkedList extractColoredChildrenExceptOne(PQRNode newAncestor, ChildrenDoublyLinkedList childList,ListNode excluded )
	{
		ChildrenDoublyLinkedList children = new ChildrenDoublyLinkedList();
		ListNode aux = head;
		ListNode iterator = startIteration();
		while (aux != null)
		{	
			ListNode auxchild = (ListNode)aux.getData();
			// removes the children 
			if (auxchild != excluded)
			{
				childList.removeNode(auxchild);
				((Node)auxchild.getData()).setParent(newAncestor);
				children.addNodeLast(auxchild);
				aux = aux.getNext(iterator);
			}
			// removes the excluded from the colored children list
			else
			{
				ListNode aux2 = aux.getNext(iterator);
				if (aux2 != null) 
				{
					if (aux.getDir1() == null || aux.getDir2() == null) // it's the head
					{
						iterator.setData(aux2); //iterator will points to the next
					}
					else
					{
						if (aux.getDir1() != aux2)
						{
							//iterator points to previous to aux2
							iterator.setData(aux.getDir1());
						}
						else
						{
							//iterator points to previous to aux2
							iterator.setData(aux.getDir2());
						}
					}
						
				}
				removeNode(aux);
				aux = aux2;
			}
		}
		return children;
	}
	
	/**
	 * Removes the first node from the list
	 * which has a specified data.
	 * @param n The data of the node to be removed.
	 */
	public void removeColoredChild(ListNode n)
	{
		ListNode aux = head;
		ListNode iterator = startIteration();
		while (aux.getData() != n)
		{
			aux = aux.getNext(iterator);
		}
		removeNode(aux);
	}
	
	
}
