package pqrtree;

/**
 * A <code>DoublyLinkedList</code> is a
 * doubly linked list with nodes that don't have
 * notion of next and previous, just two directions.
 * Each position in the list is composed by a
 * node.
 * The nodes should points to two neighbors
 * nodes in its chain, and to another object, the stored
 * data.
 * 
 * @version 0.1
 * @author Bruno Dilly
 */
public class DoublyLinkedList {
	protected int length;
	protected ListNode head, tail;
	
	/**
	 * Instanciates a new empty doubly linked 
	 * linked list.
	 */
	public DoublyLinkedList()
	{
		length = 0;
		head = null;
		tail = null;
	}
	
	/**
	 * Gets the length of the list.
	 * @return The length of the list.
	 */
	public int size()
	{
		return length;
	}
	
	/**
	 * Simply reverse the list, so the first node
	 * becomes the last, the last become the first,
	 * and so on...
	 */
	public void reverse()
	{
		ListNode aux;
		aux = head;
		head = tail;
		tail = aux;
	}

	/**
	 * Adds an already instantiated list node 
	 * at the end of the list.
	 * @param n A list node to be added to the end of the list.
	 */
	public void addNodeLast(ListNode n)
	{
		length += 1;
		if (tail != null)
		{
			if (tail.getDir1() == null)
			{
				tail.setDir1(n);
			}
			else
			{
				tail.setDir2(n);
			}
			n.setDir1(tail);
			tail = n;
		}
		else //tail == null
		{
			head = tail = n;
			n.setDir1(null);
		}
		n.setDir2(null);
	}
	
	/**
	 * Instantiates a list node, 
	 * then add it at the end of the list.
	 * @param o Tree or list node to be stored in
	 * the list.
	 * @return The list node created.
	 */
	public ListNode addLast(Object o)
	{
		ListNode n = new ListNode(o);
		addNodeLast(n);
		return n;
	}
	

	/**
	 * Adds an already instantiated list node 
	 * to the beginning of the list.
	 * @param n A list node to be added to the beginning of the list.
	 */
	public void addNodeFirst(ListNode n)
	{
		length += 1;
		if (head != null)
		{
			if (head.getDir1() == null)
			{
				head.setDir1(n);
			}
			else
			{
				head.setDir2(n);
			}
			n.setDir1(head);
			head = n;
		}
		else //head == null
		{
			head = tail = n;
			n.setDir1(null);
		}
		n.setDir2(null);
	}
	
	/**
	 * Instantiates a list node, 
	 * then add it to the beginnig of the list.
	 * @param o Tree or list node to be stored in
	 * the list.
	 * @return The list node created.
	 */
	public ListNode addFirst(Object o)
	{
		ListNode n = new ListNode(o);
		addNodeFirst(n);
		return n;
	}
	
	/**
	 * Instantiates a list node, 
	 * then add it in the same position of another one,
	 * that will be removed from the list.
	 * @param o The object to be stored into a new list node.
	 * @param old The node to be removed.
	 * @return The node which stores the object o.
	 */
	public ListNode addPos(Object o, ListNode old)
	{
		ListNode n = new ListNode(o);
		appendPos(n, n, old, old.getDir1(), old.getDir2(), 0);
		return n;
	}
	
	/**
	 * Gets the last node of the list.
	 * @return The last node.
	 */
	public ListNode getTail()
	{
		return tail;
	}
	
	/**
	 * Gets first node of the list.
	 * @return The first node.
	 */
	public ListNode getHead()
	{
		return head;
	}
	
	/**
	 * Sets the first node of the list.
	 * @param newHead The first node.
	 */
	public void setHead(ListNode newHead)
	{
		head = newHead;
	}
	
	/**
	 * Sets the last node of the list.
	 * @param newTail The last node.
	 */
	public void setTail(ListNode newTail)
	{
		tail = newTail;
	}
	
	/**
	 * Increments the length of the list.
	 * @param i Value to be added to the length.
	 */
	public void incSize(int i)
	{
		length += i;
	}
	
	/**
	 * Removes all the nodes.
	 */
	public void removeAll()	
	{ 
		head = tail = null;
		length = 0;
	}
	
	/**
	 * Removes a specific node from the list.
	 * @param v Node to be removed.
	 */
	public void removeNode(ListNode v)
	{
		if (head == v)
		{
			if (v.getDir1() != null)
			{
				head = v.getDir1();
			}
			else if (v.getDir2() != null)
			{
				head = v.getDir2();
			}
			else
			{
				head = null;
			}
		}
		if (tail==v)
		{
			if (v.getDir1() != null)
			{
				tail = v.getDir1();
			}
			else if (v.getDir2() != null)
			{
				tail = v.getDir2();
			}
			else
			{
				tail = null;
			}
		}
		v.removeMe();
		length--;
	}
	
	/**
	 * Appends another list to the beginning of
	 * the list.
	 * @param l List to be appended.
	 */
	public void appendFirst(DoublyLinkedList l)
	{
		length += l.size();
		// if list l is not empty 
		if (l.getHead() != null)
		{
			if ( head != null) // and this list neither
			{
				if (l.getTail().getDir1() == null)
				{
					l.getTail().setDir1(head);
				}
				else
				{
					l.getTail().setDir2(head);
				}
				
				if (head.getDir1()==null)
				{
					head.setDir1(l.getTail());
				}
				else
				{
					head.setDir2(l.getTail());
				}
			}
			head = l.getHead();
		}
	}
	
	/**
	 * Appends another list to an specific
	 * position of the list, excluding the node
	 * that ocupies that position.
	 * @param n1 The head or the tail of another list.
	 * The node that will be linked to the node
	 * <code>pos1</code>.
	 * @param n2 The head or the tail of another list.
	 * The node that will be linked to the node
	 * <code>pos2</code>.
	 * @param position The node that ocupies the position.
	 * @param pos1 A <code>position</code>'s neighbor.
	 * The node that will be linked to the node
	 * <code>n1</code>.
	 * @param pos2 A <code>position</code>'s neighbor.
	 * The node that will be linked to the node
	 * <code>n2</code>.
	 * @param deltaLength Value to be added to the length.
	 */
	public void appendPos(ListNode n1, ListNode n2, ListNode position, ListNode pos1, ListNode pos2, int deltaLength)
	{
		incSize(deltaLength);
		if (n1.getDir1() == null)
		{
			n1.setDir1(pos1);
		}
		else
		{
			n1.setDir2(pos1);
		}
		if (n2.getDir1() == null)
		{
			n2.setDir1(pos2);
		}
		else
		{
			n2.setDir2(pos2);
		}
		
		if (pos1 != null)
		{
			if (pos1.getDir1() == position)
			{
				pos1.setDir1(n1);
			}
			else
			{
				pos1.setDir2(n1);
			}
		}
		if (pos2 != null)
		{
			if (pos2.getDir1() == position)
			{
				pos2.setDir1(n2);
			}
			else
			{
				pos2.setDir2(n2);
			}
		}
		if (head == position)
		{
			setHead(n1);
		}
		if (tail == position)
		{
			setTail(n2);
		}	
	}
	
	/**
	 * Creates a new list node that stores the first
	 * node of the list. It's useful to make loops
	 * to search for nodes.
	 * @return A node that stores the first node
	 * of the list.
	 */
	public ListNode startIteration()
	{
		ListNode iterator = new ListNode(head);
		return iterator;
	}
	
}
