package pqrtree;

import java.awt.Color;
import java.awt.Graphics;

/**
 * A <code>node</code> object can be
 * of four types: P, Q, R - represented
 * by the <code>PQRNode</code> class - or leaves -
 * represented by the <code>Leaf</code> class.
 * 
 * @version 0.1
 * @author Bruno Dilly
 * @see pqrtree.PQRNode
 * @see pqrtree.Leaf
 */
public class Node {
	private boolean visited;
	
	private int color;
	protected boolean leaf;
	protected ListNode ancestorListPosition;
	// just for the nodes that point directly the parent
	protected PQRNode ancestor;
	// just q and r children nodes have ancestor reference
	// it is necessary to implement disjoint set's structure
	protected Node ancestorReference;
	
	protected static final int TYPE_PNODE = 0;
	protected static final int TYPE_QNODE = 1;
	protected static final int TYPE_RNODE = 2;
	
	protected static final int COLOR_WHITE = 0;
	protected static final int COLOR_GRAY = 1;
	protected static final int COLOR_BLACK = 2;
	
	// GUI related
	public static final int DRAW_SIZE = 12;
	public static final int DRAW_BOUNDARY_SIZE = 12;
	public static final int DRAW_CONNECTOR_SIZE = 24;
	public static final Color DRAW_NORMAL_COLOR = Color.black;
	public static final Color DRAW_BLACK_COLOR = Color.black;
	public static final Color DRAW_GRAY_COLOR = Color.gray;
	public static final Color DRAW_WHITE_COLOR = Color.white;
	private int subLeafCount; // number of leaf nodes under this node, field used for drawing the tree.
	private int depth; // greatest depth of node from leaf node, field used for drawing the tree.
	
	
	/**
	 * Creates a new PQRTree's node, linked to it's ancestor
	 * @param ancestor The internal node from which the new node descend
	 */
	public Node(PQRNode ancestor)
	{
		visited = false;
		color = COLOR_WHITE;
		leaf = false;
		ancestorListPosition = null;
		setParent(ancestor);
	}
	
	/**
	 * Verifies if the node was already visited.
	 * @return <code>true</code>, if the node was 
	 * visited, or <code>false</code>, if it wasn't.
	 */
	public boolean wasVisited() 
	{
		return visited;
	}
	
	/**
	 * Sets the visited state of the node.
	 * @param visited Reports if the node was visited.
	 */
	public void visited(boolean visited)
	{
		this.visited = visited;
	}
	
	/**
	 * Gets the ancestor direct link, if it exists.
	 * It just exists if the node is child of a P node.
	 * @return The ancestor, if it's a P node, or null.
	 */
	public PQRNode getAncestor()
	{
		return ancestor;
	}
	
	/**
	 * Sets the ancestor direct link.
	 * @param ancestor The ancestor, if it's a P node, or null.
	 */
	public void setAncestor(PQRNode ancestor)
	{
		this.ancestor = ancestor;
	}
	
	/**
	 * Gets the ancestor reference link, if it exists.
	 * It just exists if the node is child of a Q or R node.
	 * It's the attribute necessary to implements disjoint sets.
	 * @return The ancestor reference, that should be a brother node
	 * or maybe itself.
	 */
	public Node getAncestorReference()
	{
		return ancestorReference;
	}
	
	/**
	 * Sets the ancestor direct link.
	 * @param n The ancestor reference. Must
	 * be a brother, itself, or null.
	 */
	public void setAncestorReference(Node n)
	{
		ancestorReference = n;
	}
	
	/**
	 * Finds the imediate ancestor of the node.
	 * It can be child of P, Q or R node. The parent
	 * will be find.
	 * @return The imediate ancestor.
	 * @see #getAncestor()
	 * @see #getAncestorReference()
	 */
	public PQRNode getParent()
	{
		if (ancestorReference == null)
		{
			return ancestor;
		}
		else // it must have ancestorReference 
		{
			return findLeader().getAncestor();
		}
	}
	
	/**
	 * Sets ancestor direct link and
	 * ancestor reference to make possible
	 * finding the parent.
	 * @param p The parent.
	 * @see #setAncestor(PQRNode)
	 * @see #setAncestorReference(Node)
	 */
	public void setParent(PQRNode p)
	{
		if (p == null)
		{
			setAncestor(null);
			setAncestorReference(null);
		}
		else
		{
			switch (p.getType())
			{
			case TYPE_PNODE:
				setAncestor(p);
				setAncestorReference(null);
				break;
			case TYPE_QNODE:
			case TYPE_RNODE:
				if (p.getChildren().size() > 0)
				{
					setAncestorReference((Node)p.getFirstChild().getData());
					setAncestor(null);
				}
				else
				{
					setAncestor(p);
					setAncestorReference(this);
				}
				break;
			}
		}
	}
	
	/**
	 * Returns the color of the node.
	 * It can be, black, gray or white.
	 * @return Color of the node
	 */
	public int getColor() 
	{ 
		return color;
	}
	
	/**
	 * Sets the color of the node.
	 * It can be, black, gray or white.
	 * @param color Color black, gray or white.
	 */
	public void setColor(int color)
	{
		this.color = color;
	}
	
	/**
	 * Verifies if the node is a leaf.
	 * @return <code>true</code>, if the node is a 
	 * leaf, or <code>false</code>, if it isn't.
	 */
	public boolean isLeaf() 
	{
		return leaf;
	}
	
	/**
	 * Gets the node of the children list
	 * of the parent that points to the node.
	 * @return The list node that points to the node.
	 */
	public ListNode getAncestorListPosition()
	{
		return ancestorListPosition;
	}
	
	/**
	 * Sets the node of the children list
	 * of the parent that points to the node.
	 * @param n The list node.
	 */
	public void setAncestorListPosition(ListNode n)
	{
		ancestorListPosition = n;
	}
	
	/**
	 * Finds the representative of the disjoint set, implementing
	 * the shrink path heuristics
	 * @return The representative of the disjoint set
	 */
	public Node findLeader()
	{
		if (ancestorReference != this)
		{
			ancestorReference = ancestorReference.findLeader();
			return ancestorReference;
		}
		else
		{
			return this;
		}
	}
	
	/**
	 * Join two node children of the same node.
	 * It implements the balancing disjoint sets heuristic.
	 * @param v One node
	 * @param sizeV Weight of the first node
	 * @param p Another node
	 * @param sizeP Weight of the second node
	 * @return The node choosen to be the representative.
	 */
	public Node union (Node v, int sizeV, Node p, int sizeP) 
	{
		if (sizeV >= sizeP)
		{
			p.setAncestorReference(v);
			p.setAncestor(null);
			return v;
		}
		
		else
		{
			v.setAncestorReference(p);
			v.setAncestor(null);
			return p;
		}
	}
	
	/**
	 * Draws this node to the gui, using the provided <code>Graphics</code>
	 * object, and left	and top bounding coordinates for drawing, 
	 * and a reference to the parent node,
	 * wich allows a check for wheter his node's parent reference is valid.
	 * @param aPen The pen, graphics object.
	 * @param left Left bounding coordinates.
	 * @param top Top bounding coordinates.
	 * @param theParent The node's parent.
	 * @throws Exception
	 */
	public void drawNode(Graphics aPen, int left, int top, PQRNode theParent) throws Exception
	{
		int drawPositionY = top;
		int drawPositionX = -1;
		
		aPen.setColor(DRAW_NORMAL_COLOR);
		if (subLeafCount == 0)
		{
			drawPositionX = left + (DRAW_BOUNDARY_SIZE/2);
		}
		else
		{
			drawPositionX = left + subLeafCount*((DRAW_SIZE+DRAW_BOUNDARY_SIZE)/2) - (DRAW_BOUNDARY_SIZE/2);
		}
		
		// recursively draw any child nodes, providing them with a subset of the left and right
		// bounding coordinates of this node.
		if (isLeaf())
		{
			aPen.drawOval( drawPositionX, drawPositionY,
					DRAW_SIZE, DRAW_SIZE );
			aPen.drawString(((Leaf)this).getData().toString(), drawPositionX, drawPositionY+DRAW_SIZE+15);
			if ( this.getColor() == COLOR_BLACK)
			{
				aPen.setColor(DRAW_BLACK_COLOR);
				aPen.fillOval( drawPositionX, drawPositionY,
						DRAW_SIZE, DRAW_SIZE );
			}
		}
		if ( !isLeaf() && ((PQRNode)this).getChildren().size() > 0)
		{
			Node childNode;
			int childDrawPositionLeft = left;
			int childDrawPositionX = 0;
			ListNode iterator = ((PQRNode)this).getChildren().startIteration();
			ListNode childListNode = ((PQRNode)this).getChildren().getHead();
			
			while (childListNode != null )
			{
				childNode = (Node) childListNode.getData();
				
				if (childNode.subLeafCount == 0)
				{
					childDrawPositionX = childDrawPositionLeft + ((DRAW_SIZE+DRAW_BOUNDARY_SIZE)/2);
				}
				else
				{
					childDrawPositionX = childDrawPositionLeft + childNode.subLeafCount*((DRAW_SIZE+DRAW_BOUNDARY_SIZE)/2);
				}
				
				if ( ((PQRNode)this).getType() == TYPE_PNODE | ((PQRNode)this).getType() == TYPE_RNODE )
				{
					aPen.drawLine( drawPositionX+DRAW_SIZE/2, drawPositionY+DRAW_SIZE,
							childDrawPositionX, drawPositionY+DRAW_SIZE+DRAW_CONNECTOR_SIZE );
				}
				
				else if ( ((PQRNode)this).getType() == TYPE_QNODE )
				{
					aPen.drawLine( childDrawPositionX, drawPositionY+DRAW_SIZE,
							childDrawPositionX, drawPositionY+DRAW_SIZE+DRAW_CONNECTOR_SIZE );
				}      
				
				childNode.drawNode(aPen, childDrawPositionLeft, drawPositionY+DRAW_SIZE+DRAW_CONNECTOR_SIZE, (PQRNode)this);
				if ( childNode.subLeafCount == 0 )
				{
					childDrawPositionLeft+= (DRAW_SIZE+DRAW_BOUNDARY_SIZE);
				}
				else
				{
					childDrawPositionLeft+= childNode.subLeafCount*(DRAW_SIZE+DRAW_BOUNDARY_SIZE);
				}
				childListNode = childListNode.getNext(iterator);
			}
		}
		
		if ( !isLeaf())
		{
			
			if ( ((PQRNode)this).getType() == TYPE_PNODE )
			{
				if ( this.getColor() == COLOR_BLACK )
				{
					aPen.setColor(DRAW_BLACK_COLOR);
					aPen.fillOval( drawPositionX, drawPositionY,
							DRAW_SIZE, DRAW_SIZE );
				}
				else if ( this.getColor() == COLOR_GRAY )
				{
					aPen.setColor(DRAW_GRAY_COLOR);
					aPen.fillOval( drawPositionX, drawPositionY,
							DRAW_SIZE, DRAW_SIZE );
				}
				aPen.drawOval( drawPositionX, drawPositionY,
						DRAW_SIZE, DRAW_SIZE );
				aPen.setColor(DRAW_NORMAL_COLOR);   
			}
			else if ( ((PQRNode)this).getType() == TYPE_QNODE )
			{
				if ( this.getColor() == COLOR_BLACK )
				{
					aPen.setColor(DRAW_BLACK_COLOR);
					aPen.fillRect( left+(DRAW_BOUNDARY_SIZE/2), drawPositionY,
							(DRAW_SIZE+DRAW_BOUNDARY_SIZE)*subLeafCount-DRAW_BOUNDARY_SIZE, DRAW_SIZE );
				}
				else if ( this.getColor() == COLOR_GRAY )
				{
					aPen.setColor(DRAW_GRAY_COLOR);
					aPen.fillRect( left+(DRAW_BOUNDARY_SIZE/2), drawPositionY,
							(DRAW_SIZE+DRAW_BOUNDARY_SIZE)*subLeafCount-DRAW_BOUNDARY_SIZE, DRAW_SIZE );
				}
				aPen.drawRect( left+(DRAW_BOUNDARY_SIZE/2), drawPositionY,
						(DRAW_SIZE+DRAW_BOUNDARY_SIZE)*subLeafCount-DRAW_BOUNDARY_SIZE, DRAW_SIZE );
				aPen.setColor(DRAW_NORMAL_COLOR);
			}
			else if ( ((PQRNode)this).getType() == TYPE_RNODE )
			{
				if ( this.getColor() == COLOR_BLACK )
				{
					aPen.setColor(DRAW_BLACK_COLOR);
					aPen.fillOval( drawPositionX, drawPositionY,
							DRAW_SIZE, DRAW_SIZE );
					aPen.setColor(DRAW_WHITE_COLOR);
					aPen.drawString("R", drawPositionX+3, drawPositionY+DRAW_SIZE);
					aPen.setColor(DRAW_BLACK_COLOR);
				}
				else if ( this.getColor() == COLOR_GRAY )
				{
					aPen.setColor(DRAW_GRAY_COLOR);
					aPen.fillOval( drawPositionX, drawPositionY,
							DRAW_SIZE, DRAW_SIZE );
					aPen.setColor(DRAW_BLACK_COLOR);
					aPen.drawString("R", drawPositionX+3, drawPositionY+DRAW_SIZE);
					aPen.setColor(DRAW_GRAY_COLOR);
				}
				else
				{
					aPen.drawString("R", drawPositionX+3, drawPositionY+DRAW_SIZE);
				}
				aPen.drawOval( drawPositionX, drawPositionY,
						DRAW_SIZE, DRAW_SIZE );
				aPen.setColor(DRAW_NORMAL_COLOR);   
			}
		}
	}
	
	/**
	 * Returns a count of the number of leaves stored underneath this node.
	 * This method also calculates the greatest depth of this node from a leaf node.
	 * @param parent_depth Greatest parent depth.
	 * @return Count of the number of leaves.
	 * @throws Exception
	 */
	public int countSubLeaves(int parent_depth) throws Exception
	{
		subLeafCount = 0;
		depth = parent_depth+1;
		int tempDepth = depth;
		if ( !isLeaf() && ((PQRNode)this).getChildren().size() > 0 )
		{
			Node childNode;
			ListNode iterator = ((PQRNode)this).getChildren().startIteration();
			ListNode childListNode = ((PQRNode)this).getChildren().getHead();
			
			while (childListNode != null )
			{
				childNode = (Node) childListNode.getData();
				subLeafCount += childNode.countSubLeaves(tempDepth);
				if (childNode.depth > depth)
				{
					depth = childNode.depth;
				}
			childListNode = childListNode.getNext(iterator);
			}
			return subLeafCount;
		}
		else
		{
			return 1;
		}
	}
	
	/**
	 * Gets the greatest node depth.
	 * @return Greatest depth.
	 */
	public int getDepth()
	{ 
		return depth; 
	}
}
