package pqrtree;

/**
 * A <code>PQRNode</code> object is an intern
 * node of PQRTree. It can be type P, Q or R.
 * Every P Node has at least 2 children, that
 * can be arbitrary permutated.
 * Every Q Node has at least 3 children, that
 * can be reversed.
 * Every R Node has at least 3 children, that
 * can be arbitrary permutated. The R nodes
 * encapsulate subsets where the consecutive 
 * ones proprety fails.
 * 
 * @version 0.1
 * @author Bruno Dilly
 * @see pqrtree.Node
 */
public class PQRNode extends Node {
	private int type;
	
	private int pertinentPaths;
	private int pertinentLeaves;
	private boolean deleted = false;
	
	private ChildrenDoublyLinkedList children;
	private ColoredDoublyLinkedList blackChildren, grayChildren;
	
	/**
	 * Creates a new internal node of the PQRTree, type P, Q or R,
	 * linked to it's ancestor.
	 * @param type The type of the internal node. Could be a P, Q or R node.
	 * @param ancestor The internal node from which the new node descend.
	 * @see Node#Node(PQRNode)
	 */
	public PQRNode(int type, PQRNode ancestor)
	{
		super(ancestor);
		this.type = type;
		pertinentPaths = 0;
		pertinentLeaves = 0;
		children = new ChildrenDoublyLinkedList();
		blackChildren = new ColoredDoublyLinkedList();
		grayChildren = new ColoredDoublyLinkedList();
	}
	
	/**
	 * Returns the type of the node.
	 * @return The type of the node: P, Q or R node.
	 */
	public int getType()  
	{
		return type;
	}
	
	/**
	 * Sets the type of the node. It's necessary when a explicit change is required.
	 * @param type The type of the node: P, Q or R node.
	 */
	public void setType(int type)
	{
		this.type = type;
	}
	
	/**
	 * Increments the pertinent paths counter by one.
	 */
	public void incPertinentPaths() 
	{ 
		pertinentPaths++;
	}
	
	/**
	 * Decrements the pertinent paths counter by one.
	 */
	public void decPertinentPaths() 
	{ 
		pertinentPaths--;
	}
	
	/**
	 * Sets the pertinent paths counter to zero.
	 */
	public void zeroPertinentPaths() 
	{ 
		pertinentPaths = 0;
	}
	
	/**
	 * Sets the pertinent paths counter to a new value.
	 * @param newValue The new pertinent paths value.
	 */
	public void setPertinentPaths(int newValue)  
	{
		pertinentPaths = newValue;
	}
	
	/**
	 * Gets the pertinent paths counter value.
	 * @return The pertinent paths counter value.
	 */
	public int getPertinentPaths()
	{
		return pertinentPaths;
	}
	
	/**
	 * Increments the pertinent leaves counter by a integer value.
	 * @param i The value of pertinent leaves to be added to 
	 * the pertinent leaves counter.
	 */
	public void incPertinentLeaves(int i) 
	{
		pertinentLeaves += i;
	}
	
	/**
	 * Sets the pertinent leaves counter to zero.
	 */
	public void zeroPertinentLeaves() 
	{ 
		pertinentLeaves = 0;
	}
	
	/**
	 * Sets the pertinent leaves counter to a new value.
	 * @param newValue The new pertinent leaves value.
	 */
	public void setPertinentLeaves(int newValue) 
	{
		pertinentLeaves = newValue;
	}
	
	/**
	 * Gets the pertinent leaves counter value.
	 * @return The pertinent leaves counter value.
	 */
	public int getPertinentLeaves()
	{ 
		return pertinentLeaves;
	}
	
	/**
	 * Gets the children's list and returns it.
	 * @return The children's list.
	 */
	public ChildrenDoublyLinkedList getChildren()
	{
		return children;
	}
	
	/**
	 * Gets the black children's list and returns it.
	 * @return The black children's list.
	 */
	public ColoredDoublyLinkedList getBlackChildren()
	{
		return blackChildren;
	}
	
	/**
	 * Gets the gray children's list and returns it.
	 * @return The gray children's list.
	 */
	public ColoredDoublyLinkedList getGrayChildren(){return grayChildren;}
	
	/**
	 * Changes the children's list by another one.
	 * @param newList A new children's list.
	 */
	public void setChildren(ChildrenDoublyLinkedList newList)
	{
		children = newList;
	}
	
	/**
	 * Changes the black children's list by another one.
	 * @param newList A new black children's list.
	 */
	public void setBlackChildren(ColoredDoublyLinkedList newList)
	{
		blackChildren = newList;
	}
	
	/**
	 * Changes the gray children's list by another one.
	 * @param newList A new gray children's list.
	 */
	public void setGrayChildren(ColoredDoublyLinkedList newList)
	{
		grayChildren = newList;
	}
	
	/**
	 * Verifies if all black children are consecutive.
	 * @return <code>true</code>, if all black children are consecutive, or <code>false</code>, 
	 * if they aren't.
	 */
	public boolean areBlackChildrenConsecutive()
	{
		return children.areAllColoredChildrenConsecutive(COLOR_BLACK);
	}
	
	/**
	 * Verifies if all children are black.
	 * @return <code>true</code>, if all children are black, or <code>false</code>, 
	 * if they aren't.
	 */
	public boolean areAllChildrenBlack()
	{
		//it just compares the chidren and black chidren lists length
		return (blackChildren.size() == children.size());
	}
	
	/**
	 * Adds a node to the end of the children's list.
	 * @param child Node to be added to the children's list.
	 */
	public void addChild(Node child)
	{
		ListNode position = children.addLast(child);
		child.setAncestorListPosition(position);
	}
	
	/**
	 * Adds a child node to the end of the black children's list.
	 * It means that this child will be considered black by the ancestor.
	 * @param child Node to be considered black by the ancestor.
	 */
	public void addBlackChild(Node child)
	{
		blackChildren.addLast(child.getAncestorListPosition());
	}
	
	/**
	 * Adds a child node to the end of the gray children's list.
	 * It means that this child will be considered gray by the ancestor.
	 * @param child Node to be considered gray by the ancestor.
	 */
	public void addGrayChild(Node child)
	{
		grayChildren.addLast(child.getAncestorListPosition());
	}
	
	/**
	 * Empties the black children's list.
	 */
	public void clearBlackChildren()
	{
		blackChildren = new ColoredDoublyLinkedList();
	}
	
	/**
	 * Empties the gray children's list.
	 */
	public void clearGrayChildren()
	{
		grayChildren = new ColoredDoublyLinkedList();
	}
	
	/**
	 * Empties the children's list.
	 */
	public void clearChildren()
	{
		children = new ChildrenDoublyLinkedList();
	}
	
	/**
	 * Verifies if the node has at least one gray child.
	 * @return <code>true</code>, if it has at least onde gray child, or <code>false</code>, 
	 * if it doesn't.
	 */
	public boolean hasGrayChildren()
	{
		return grayChildren.size() > 0;
	} 
	
	/**
	 * Gets a gray child node, and returns it. It isn't removed.
	 * @return A gray child.
	 */
	public PQRNode getGrayChild()
	{
		ListNode iterator = getGrayChildren().startIteration();
		ListNode aux = getGrayChildren().getHead();
		while (aux != null)
		{
			if (!((PQRNode)((ListNode)aux.getData()).getData()).wasDeleted())
			{
				break;
			}
			aux = aux.getNext(iterator);
		}
		return (PQRNode)((ListNode)(aux.getData())).getData();
	}
	
	/**
	 * Gets the first child, and returns it. It isn't removed.
	 * @return The first child.
	 */
	public ListNode getFirstChild()
	{
		return children.getHead();
	}
	
	/**
	 * Gets the last child, and returns it. It isn't removed.
	 * @return The last child.
	 */
	public ListNode getLastChild()
	{
		return children.getTail();
	}
	
	
	/**
	 * Substitutes the node by another one. So, the ancestor, the ancestor list position,
	 * and the reference ancestor is hold. All the other attributes are changed by the
	 * other node ones.
	 * @param v The node that will substitutes this one. v must be a Q or R node.
	 */
	public void substitutedBy(PQRNode v)
	{
		setType(v.getType());
		setBlackChildren(v.getBlackChildren());
		setGrayChildren(v.getGrayChildren());
		setChildren(v.getChildren());
		setPertinentPaths(v.getPertinentPaths());
		setPertinentLeaves(v.getPertinentLeaves());
		//the children must point to this node
		((Node)v.getFirstChild().getData()).findLeader().setAncestor(this);
	}
	
	/**
	 * Verifies if the node was deleted
	 * @return <code>true</code>, if the was deleted,  
	 * or <code>false</code>, if it wasn't.
	 */
	public boolean wasDeleted()
	{
		return deleted;
	}
	
	/**
	 * Sets the node as deleted.
	 */
	public void setDeleted()
	{
		this.deleted = true;
	}
}
