package pqrtree;
import java.util.LinkedList;
import java.awt.*;
import java.util.Vector;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.Condition;
import pqrtreeapplet.*;

/**
 * A PQR Tree over a set U is a rooted tree with four types of
 * nodes - P, Q, R and leaves.
 * PQR Tree is a compact way of storing and manipulating all the permutantion
 * of n elements that keep consecutive the elements in certaing given sets.
 * And encapsulates subsets where the consecutive ones fails by R nodes.
 * 
 * @version 0.1
 * @author Bruno Dilly
 *
 */
public class PQRTree implements Runnable {
	private PQRNode root; // root of the PQR tree
	private Leaf[] leaves; // leaves of the PQR tree
	
	private static final int TYPE_PNODE = 0;
	private static final int TYPE_QNODE = 1;
	private static final int TYPE_RNODE = 2;
	
	private static final int COLOR_WHITE = 0;
	private static final int COLOR_GRAY = 1;
	private static final int COLOR_BLACK = 2;
	private int[] s; //constraints
	
	// GUI related
	private int drawWidth;
	private int drawHeight;
	private boolean hasChanged;
	private boolean doneReduction;
	private String reduceString;
	private String operationsString;
	private boolean stepByStep;
	private boolean setToBeAdded;
	private boolean painted;
	public Lock addSetLock;
	public Condition canAddSet, canBePainted, nextStep;
	public PQRController controller;
	
	/**
	 * Creates a universal PQRTree. It's a tree with a PNode with all the 
	 * elements of the universe set as it's leaves.
	 * @param u The universe set. Will be represented as leaves.
	 */
	public PQRTree(Object[] u)
	{
		doneReduction = true;
		root = new PQRNode(TYPE_PNODE, null);
		leaves = new Leaf[u.length];
		int i = 0;
		try
		{
			for ( i=0; i<u.length; i++ )
			{
				Leaf aLeaf = new Leaf(u[i], root);
				root.addChild(aLeaf);
				leaves[i] = aLeaf;
			}
		}
		catch (OutOfMemoryError e)
		{
			System.out.println("Insufficient memory while creating leaf node #" + i);
			throw e;
		}
		hasChanged = true;
	}
	
	/**
	 * Creates a universal PQRTree. It's a tree with a PNode with all the 
	 * elements of the universe set as it's leaves.
	 * @param u The universe set. Will be represented as leaves.
	 * @param controller Graphics controller.
	 * @param addSetLock A lock to works with some conditions.
	 * @param canAddSet Condition to adds a set to the tree.
	 * @param canBePainted Condition to paints the tree.
	 * @param nextStep Condition to executes the next step.
	 */
	public PQRTree(Object[] u, PQRController controller, Lock addSetLock,
			Condition canAddSet, Condition canBePainted, Condition nextStep)
	{
		this(u);
		setLocks(addSetLock, canAddSet, canBePainted, nextStep);
		this.controller = controller;
		
	}
	
	/**
	 * Creates a universal PQRTree. It's a tree with a PNode with all the
	 * elements of the universe set as it's leaves. For each set
	 * of the collection, it adds the set. It's a online algorithm.
	 * @param u The universe set. Will be represented as leaves.
	 * @param c The collection. A set of sets.
	 * @see #PQRTree(Object[])
	 */
	public PQRTree(Object[] u, int[][] c)
	{
		this(u);
		int i = 0;
		// add all sets of c collection to pqr tree
		for (i=0; i<c.length; i++ )
		{
			addSet(c[i]);
		}
		
	}
	
	/**
	 * Runs the thread.
	 */
	public void run ()
	{
		while(true)
		{
			addSetLock.lock();
			try
			{
				while(!setToBeAdded)
				{
					canAddSet.await();
				}
				setToBeAdded = false;
				addSet();
				controller.setToBePainted = true;
				canBePainted.signalAll();
				addSetLock.unlock();
			}
			catch(InterruptedException e)
			{
				//System.out.println("Error :" + e.toString());
			}
		}
	}
	
	/**
	 * Sets the constraints.
	 * @param s The constraints to set.
	 */
	public void setConstraints(int[] s)
	{
		this.s = s;
		setToBeAdded = true;
	}
	
	/**
	 * Changes the tree to keep consecutive 
	 * the elements of the set previously added, if it's possible.
	 */
	public void addSet()
	{
		doneReduction = false;
		operationsString = "Operations: { ";
		if (s.length > 1)
		{
			step();
			operationsString += "CL, ";
			PQRNode w = colorTree(s); //find the LCA
			w = reduceTree(w); // update LCA
			adjustLCA(w);
			operationsString += "UT ";
			uncolorTree(s);
		}
		hasChanged = true;
		operationsString += "}\n";
		setReduceString(s);
		doneReduction = true;
	}
	
	/**
	 * Changes the tree to keep consecutive 
	 * the elements of the set <code>s</code>, if it's possible.
	 * @param s Set of elements that must to be consecutive.
	 */
	public void addSet(int[] s)
	{
		this.s = s;
		setToBeAdded = true;
	}
	

	/**
	 * Colors black the leaves corresponding to the elements of
	 * <code>s</s>, colors black, gray or keeps white the internal
	 * nodes, and finds the Least Commons Ancestor.
	 * @param s Set of elements that must to be consecutive.
	 * @return The Least Common Ancestor.
	 */
	private PQRNode colorTree(int[] s)
	{
		LinkedList aQueue = new LinkedList();
		Node v = null;
		PQRNode p;
		
		int gotroot = 0;
		for (int i=0; i<s.length; i++)
		{
			v = leaves[s[i]];
			v.visited(true);
			aQueue.addLast(v);
		}
		
		while (aQueue.size() + gotroot > 1)
		{
			v = (Node)aQueue.removeFirst();
			p = v.getParent();
			if (p != null) 
			{
				p.incPertinentPaths();
				if (!p.wasVisited())
				{
					p.visited(true);
					aQueue.addLast(p);
				}
			}
			else
			{
				gotroot = 1;
			}
		}

		aQueue = new LinkedList();
		for (int i=0; i<s.length; i++)
		{
			v = leaves[s[i]];
			aQueue.addLast(v);
		}
		
		while (aQueue.size() > 0) 
		{
			v = (Node)aQueue.removeFirst();
			p = v.getParent();
			
			if (v.isLeaf())
			{
				p.decPertinentPaths();
				p.incPertinentLeaves(1);
				v.setColor(COLOR_BLACK);
				p.addBlackChild(v);
				step();
			}
			else //v is an intern node, so a PQRNode
			{
				if (((PQRNode)v).getPertinentLeaves() == s.length)
				{
					break; //LCA was reached
				}
				else 
				{
					p.decPertinentPaths();
					p.incPertinentLeaves(((PQRNode)v).getPertinentLeaves());
				}
				if (((PQRNode)v).areAllChildrenBlack())
				{
					((PQRNode)v).setColor(COLOR_BLACK);
					p.addBlackChild(((PQRNode)v));
					step();
				}
				else
				{
					((PQRNode)v).setColor(COLOR_GRAY);
					p.addGrayChild(((PQRNode)v));
					step();
				}
			}
			if (p.getPertinentPaths() == 0)
			{
				aQueue.addLast(p);
			}
		}
		return (PQRNode)v; //returns the LCA
	}
	
	/**
	 * Repeatedly kills, merge, or uncolors gray nodes,
	 * sometimes also creating new nodes when needed, until no
	 * gray nodes remain in the tree.
	 * @param w Least Common Ancestor
	 * @return The Least Common Ancestor, that can be another
	 * node after the reduction.
	 */
	private PQRNode reduceTree(PQRNode w)
	{
		while (w.hasGrayChildren())
		{
			PQRNode v = w.getGrayChild();
			switch (w.getType())
			{
			case TYPE_PNODE:
				switch (v.getType())
				{
				case TYPE_PNODE:
					v = preparePNode(v);
					operationsString += "PP, ";
					step();
					changePIntoQ(v);
					operationsString += "PQ, ";
					step();
					break;
				case TYPE_QNODE:
					prepareLCA(w);
					operationsString += "PL, ";
					step();
					reverseGrayNode(v);
					operationsString += "RG, ";
					step();
					w = moveChildrenToGrayNode(w, v); // warning: update the LCA!
					operationsString += "MG, ";
					step();
					break;
				case TYPE_RNODE:
					prepareLCA(w);
					operationsString += "PL, ";
					step();
					w = moveChildrenToGrayNode(w, v); // warning: update the LCA!
					operationsString += "MG, ";
					step();
					break;
				}
				break;
			case TYPE_QNODE:
				switch (v.getType())
				{
				case TYPE_PNODE:
					v = preparePNode(v);
					operationsString += "PP, ";
					step();
					changePIntoQ(v);
					operationsString += "PQ, ";
					step();
					break;
				case TYPE_QNODE:
					reverseGrayNode(v);
					operationsString += "RG, ";
					step();
				case TYPE_RNODE:
					moveChildrenToLCA(w, v);
					operationsString += "ML, ";
					step();
					break;
				}
				break;
			case TYPE_RNODE:
				switch (v.getType())
				{
				case TYPE_PNODE:
					v = preparePNode(v);
					operationsString += "PP, ";
					step();
					changePIntoQ(v);
					operationsString += "PQ ,";
					step();
					break;
				case TYPE_QNODE:
				case TYPE_RNODE:
					moveChildrenToLCA(w, v);
					operationsString += "ML, ";
					step();
					break;
				}
				break;
			}
		}
		// When children are moved from LCA to a gray child
		// this new gray child is the new LCA
		return w;
	}	
	

	/**
	 * Adjusts the LCA according to its type.
	 * After reduction there are no gray nodes left in the tree.
	 * Therefore, all maximal black nodes are children of the LCA.
	 * @param w Least Common Ancestor
	 */
	private void adjustLCA(PQRNode w)
	{
		if (w.getType()==TYPE_PNODE)
		{
			prepareLCA(w);
			operationsString += "PL, ";
			step();
		}
		else if (w.getType()==TYPE_QNODE && !w.areBlackChildrenConsecutive())
		{
			changeQIntoR(w);
			operationsString += "QR, ";
			step();
		}
	}
	
	/**
	 * Uncolors all black nodes
	 * @param s Set of elements that must to be consecutive.
	 */
	private void uncolorTree(int[] s)
	{
		LinkedList aQueue = new LinkedList();
		Node v;
		PQRNode p;
		for (int i=0; i<s.length; i++)
		{
			v = leaves[s[i]];
			aQueue.addLast(v);
		}
		
		while (aQueue.size() > 0) 
		{
			v = (Node)aQueue.removeFirst();
			v.visited(false);
			p = v.getParent();

			if (!v.isLeaf())
			{
				if (((PQRNode)v).getPertinentLeaves() > 0)
				{
					((PQRNode)v).zeroPertinentPaths();
					((PQRNode)v).zeroPertinentLeaves();
					((PQRNode)v).clearBlackChildren();
					((PQRNode)v).clearGrayChildren();
					if (((PQRNode)v).getColor() != COLOR_WHITE)
					{
						((PQRNode)v).setColor(COLOR_WHITE);
						step();
					}
				}
			}
			else //v.isLeaf()
			{
				v.setColor(COLOR_WHITE);
				step();
			}
			if (p != null  && (p.getPertinentLeaves() > 0 || p.getPertinentPaths() > 0))
			{				
				aQueue.addLast(p);
			}		
		}
	}
	
	
	/////////////////////////////
	/// REDUCTION OPERATIONS ///
	///////////////////////////

	
	/**
	 * Groups the black children of LCA into a new P node.
	 * It's applied when the LCA <code>w</code> is a P node.
	 * @param w Least Common Ancestor
	 */
	private void prepareLCA(PQRNode w)
	{
		if (w.getBlackChildren().size() >= 2 & w.getBlackChildren().size() < w.getChildren().size())
		{
			PQRNode p = new PQRNode(TYPE_PNODE, w);
			p.setColor(COLOR_BLACK);
			// move black children from w to p
			p.setChildren(w.getBlackChildren().extractColoredChildren(p, w.getChildren()));
			p.setBlackChildren(w.getBlackChildren());
			p.setPertinentLeaves(p.getBlackChildren().size());
			w.clearBlackChildren();
			w.addChild(p);
			w.addBlackChild(p);
		}
	}
	
	/**
	 * Groups black children and white children into new P nodes.
	 * It's applied to a gray P node child of the LCA.
	 * @param v A gray P node, child of the LCA.
	 * @return Updates v node.
	 */
	private PQRNode preparePNode(PQRNode v)
	{
		//first, creates a p node with all black children, if it's necessary
		prepareLCA(v);
		//so v has at most one black child
		
		//then, creates a node with all white children in a p node, if necessary
		//if white children >= 2
		if (v.getChildren().size() - (v.getBlackChildren().size() + v.getGrayChildren().size()) >=2)
		{
			//creates the new node that will substitute v as children of v parent
			PQRNode newNode = new PQRNode(TYPE_PNODE, v.getParent());
			//copies the ancestor and ancestor reference from v
			newNode.setAncestor(v.getAncestor());
			if (v.getAncestorReference() == v) //v is the leader
			{
				newNode.setAncestorReference(newNode); //new Node is the new leader
			}
			else
			{
				newNode.setAncestorReference(v.getAncestorReference());
			}
			//it will be gray
			newNode.setColor(COLOR_GRAY);	
			//and adds the new node in the same position v was.
			ListNode position = v.getParent().getChildren().addPos(newNode, v.getAncestorListPosition());
			newNode.setAncestorListPosition(position);

			//removes v from gray children list
			v.getParent().getGrayChildren().removeColoredChild(v.getAncestorListPosition());
			//add newNode to graychildren list
			v.getParent().getGrayChildren().addFirst(newNode.getAncestorListPosition());
			v.setParent(newNode); //new node is a p node, so it can be done now
			
			// the new node gets all the gray children from v
			newNode.setChildren(v.getGrayChildren().extractColoredChildren(newNode, v.getChildren()));
			newNode.setGrayChildren(v.getGrayChildren());
			v.clearGrayChildren();
			
			//removes black child from v to newNode
			if (v.getBlackChildren().size() == 1) // v has at most one black child
			{
				Node blackNode = (Node)v.getBlackChildren().extractColoredChildren(newNode, v.getChildren()).getHead().getData();
				newNode.addChild(blackNode);
				newNode.addBlackChild(blackNode);
				v.clearBlackChildren();
			}
			
			newNode.setPertinentLeaves(v.getPertinentLeaves());
			// now the node v only has white children
			v.setColor(COLOR_WHITE);
			v.setPertinentLeaves(0);
			v.setPertinentPaths(0);
			v.visited(false);
			newNode.addChild(v);
			
			//now, the gray child of the LCA is newNode
			return newNode;
		}
		return v;
	}
	
	/**
	 * Changes a Q LCA with black children not consecutive
	 * into a R node.
	 * @param w Least Common Ancestor of type Q.
	 */
	private void changeQIntoR(PQRNode w)
	{
		w.setType(TYPE_RNODE);
	}
	
	/**
	 * Changes the type of the node from P to Q.
	 * The P node is a gray child of the LCA, with at
	 * most one black child and one white child.
	 * @param v A gray P node, child of the LCA.
	 */
	private void changePIntoQ(PQRNode v)
	{
		// sort the children
		
		// it has at most one black child
		if (v.getBlackChildren().size() == 1)
		{
			v.getChildren().removeNode((ListNode)v.getBlackChildren().getHead().getData());
			v.getChildren().addNodeFirst((ListNode)v.getBlackChildren().getHead().getData());
		}
		
		// change the type
		v.setType(TYPE_QNODE);
		
		if (((Node)v.getChildren().getHead().getData()).getColor() == COLOR_WHITE)
		{
			ListNode aux = ((Node)v.getChildren().getHead().getData()).getAncestorListPosition();
			v.getChildren().removeNode(aux);
			v.getChildren().addNodeLast(aux);
		}
		// creating the disjoint set structure - make set
		Node leader = (Node)v.getChildren().getHead().getData();
		ListNode iterator = v.getChildren().startIteration();
		ListNode nextChild = v.getChildren().getHead().getNext(iterator);
		ListNode aux;
		leader.setAncestorReference(leader);
		while (nextChild != null)
		{
			leader = v.union(leader, 1, (Node)nextChild.getData(), 0);
			//searches for the white node and moves it to the end of the list
			if (((Node)nextChild.getData()).getColor() == COLOR_WHITE)
			{
				aux = nextChild.getNext(iterator);
				if (aux != null) 
				{
					if (nextChild.getDir1() == null || nextChild.getDir2() == null) // it's the head
					{
						iterator.setData(aux); //iterator will points to the next
					}
					else
					{
						if (nextChild.getDir1() != aux)
						{
							//iterator points to previous to aux2
							iterator.setData(nextChild.getDir1());
						}
						else
						{
							//iterator points to previous to aux2
							iterator.setData(nextChild.getDir2());
						}
					}
				}
				v.getChildren().removeNode(nextChild);
				v.getChildren().addNodeLast(nextChild);
				nextChild = aux;
			}
			else
			{
				nextChild = nextChild.getNext(iterator);
			}
			
		}
		
	}
	
	/**
	 * Reverse a Q gray node, if its children isn't 
	 * in the right order.
	 * @param v Q gray node.
	 */
	private void reverseGrayNode(PQRNode v)
	{
		// if v is in right order
		if ((((Node)v.getFirstChild().getData()).getColor() == COLOR_BLACK) |
				(((Node)v.getFirstChild().getData()).getColor() == COLOR_GRAY &
						((Node)v.getLastChild().getData()).getColor() == COLOR_WHITE) |
						(((Node)v.getFirstChild().getData()).getColor() == COLOR_WHITE &
								((Node)v.getLastChild().getData()).getColor() == COLOR_WHITE))
		{
			return; 
		}
		// if v is reversed
		else
		{
			v.getChildren().reverse();
		}
	}
	
	/**
	 * A gray child of the LCA become the ancestor of
	 * its black and gray brother nodes. 
	 * @param w Least Common Ancestor type P with 
	 * at most one black child
	 * @param v A Q node in right order or a R node, 
	 * gray child of the LCA.
	 */
	private PQRNode moveChildrenToGrayNode(PQRNode w, PQRNode v)
	{
		//extract black child, if it exists
		ChildrenDoublyLinkedList blackList;
		blackList = w.getBlackChildren().extractColoredChildren(v, w.getChildren());
		// extract gray children
		ChildrenDoublyLinkedList grayList;
		grayList = w.getGrayChildren().extractColoredChildrenExceptOne(v, w.getChildren(), v.getAncestorListPosition());
		// move black and gray children to v, in the order
		// gray children, black, original v children
		v.getChildren().appendFirst(blackList);
		v.getChildren().appendFirst(grayList);
		v.getBlackChildren().appendFirst(w.getBlackChildren());
		v.getGrayChildren().appendFirst(w.getGrayChildren());
		
		// make w colored children lists consistent
		w.getBlackChildren().removeAll();
		w.getGrayChildren().removeAll();
		
		// v is painted white
		v.setColor(COLOR_WHITE);
		
		if (w.getChildren().size() == 1)
		{
			w.substitutedBy(v);
			return w; //the LCA continues to be w, because it was substituted by v
		}
		return v; // the new LCA is v
	}
	
	/**
	 * Move the children from a gray node, child of the LCA,
	 * to the LCA.
	 * @param w Least Common Ancestor type Q, in right order
	 * relative to the child <code>v</code>, or type R.
	 * @param v Gray node type Q in right order or type R. 
	 */
	private void moveChildrenToLCA(PQRNode w, PQRNode v)
	{

		//fixing the ancestor reference
		((Node)v.getFirstChild().getData()).findLeader().setAncestor(null);
		((Node)v.getFirstChild().getData()).findLeader().setAncestorReference(v);
		
		//append v's children list to w children list in the position v was
		ListNode neighbor1 = v.getAncestorListPosition().getDir1();
		ListNode neighbor2 = v.getAncestorListPosition().getDir2();
		// it's necessary to choose neighbor 1 and 2 in the right order
		// so the v's children can be added in the right order
		if (w.getFirstChild().getData() == v) // if v is the first child of w
		{
			if (neighbor2 == null) // the neighbor1 must be null
			{
				neighbor2 = neighbor1;
				neighbor1 = null;
			}
			v.getChildren().reverse(); // the lca isn't in the right order
		}
		else if (w.getLastChild().getData() == v) // if v is the last child of w
		{
			if (neighbor1 == null) // the neighbor2 must be null
			{
				neighbor1 = neighbor2;
				neighbor2 = null;
			}
		}
		else if (neighbor1 != null)
		{
			if (neighbor2 != null)
			{
				// if they are in the wrong order
				if (((Node)neighbor1.getData()).getColor() < ((Node)neighbor2.getData()).getColor())
				{
					ListNode aux = neighbor1;
					neighbor1 = neighbor2;
					neighbor2 = aux;
				}
			}
		}
		w.getChildren().appendPos(v.getFirstChild(), v.getLastChild(), v.getAncestorListPosition(), neighbor1, neighbor2, v.getChildren().size()-1);
		
		//mark v as deleted
		v.setDeleted();
		// decrement the sum of gray children
		w.getGrayChildren().incSize(-1);
		
		//append black and gray children lists to w
		w.getBlackChildren().appendFirst(v.getBlackChildren());
		w.getGrayChildren().appendFirst(v.getGrayChildren());
		
		
		if (w.getType() == TYPE_QNODE & v.getType() ==TYPE_RNODE)
		{
			w.setType(TYPE_RNODE);
		}
	}
	
	/**
	 * Finds a valid permutation given the collection.
	 * @return A valid permutation.
	 */
	public String frontier()
	{
		return frontier(root);
	}
	
	/**
	 * Creates a string representing a valid permutation
	 * recursively.
	 * @param v Node to be evaluated.
	 * @return The leaves data converted to string.
	 */
	private String frontier(Node v)
	{
		String frontier = "";
		if (v.isLeaf())
		{
			frontier = ((Leaf)v).getData().toString() + " ";
		}
		else
		{
			ListNode aux = ((PQRNode)v).getFirstChild();
			ListNode iterator = ((PQRNode)v).getChildren().startIteration();
			while (aux != null)
			{
				frontier += frontier( (Node)(aux.getData()) );
				aux = aux.getNext(iterator);
			}
		}
		return frontier;
	}
	
	/**
	 * Represents the PQR tree in pre order way.
	 * Displays intern nodes type and leaves data.
	 * @return Representation of the PQR tree.
	 */
	public String preOrder()
	{
		return preOrder(root);
	}
	
	/**
	 * Creates a string representing the tree
	 * recursively.
	 * @param v Node to be evaluated.
	 * @return The nodes hierarchy converted to string.
	 */
	private String preOrder(Node v)
	{
		String preOrder = "";
		if (v.isLeaf())
		{
			preOrder = " " + ((Leaf)v).getData().toString();
		}
		else
		{
			preOrder += "(";
				switch (((PQRNode)v).getType())
				{
				case TYPE_PNODE: preOrder+= "P"; break;
				case TYPE_QNODE: preOrder+= "Q"; break;
				case TYPE_RNODE: preOrder+= "R"; break;
				}
			ListNode aux = ((PQRNode)v).getFirstChild();
			ListNode iterator = ((PQRNode)v).getChildren().startIteration();
			while (aux != null)
			{
				preOrder += preOrder( (Node)(aux.getData()) );
				aux = aux.getNext(iterator);
			}
			preOrder += ")";
		}
		return preOrder;
	}
	
	/**
	 * Gets the leaves of the Tree
	 * @return Leaves
	 */
	public Vector getLeaves()
	{
		Vector elem = new Vector();
		for (int i=0; i<leaves.length; i++)
		{
			elem.add(leaves[i].getData());
		}
		return elem;
	}
	
	public int getWidth() 
	{ 
		return drawWidth; 
	}
	
	public int getHeight() 
	{ 
		return drawHeight; 
	}
	
	/**
	 * Prepares the PQTree to be drawn by the GUI by calculating the width and height
	 * needed to draw the tree, and calculated the width and height required to draw
	 * each node of the tree.
	 */
	public void prepareToDrawTree() throws Exception
	{
		if (hasChanged)
		{
			int depth = 0;
			int width = root.countSubLeaves(depth);
			drawWidth = width*(PQRNode.DRAW_SIZE+PQRNode.DRAW_BOUNDARY_SIZE)+50;
			drawHeight = root.getDepth()*(PQRNode.DRAW_SIZE+PQRNode.DRAW_CONNECTOR_SIZE)+50;
		}
	}
	
	/**
	 * Draws the tree to the GUI.
	 * @param aPen The pen, graphics object.
	 */
	public void drawTree(Graphics aPen) throws Exception
	{
		root.drawNode(aPen, 5, 5, null);
	}
	
	/**
	 * Returns a leaf node with the specified index relative the the order in which
	 * they were originally added when the PQTree was built.
	 * @param index The index of leaves vector.
	 * @return The leaf.
	 */
	public Node getLeafAt(int index)
	{
		return leaves[index];
	}
	
	/**
	 * Gets the reduce string. The leaves that compose the constraint.
	 * @return Reduce string.
	 */
	public String getReduceString() 
	{ 
		String aux = reduceString;
		reduceString ="";
		return aux;
	}
	
	/**
	 * Gets the operations string. All the operantions executed by
	 * the tree.
	 * @return Operations string.
	 */
	public String getOperationsString()
	{
		String aux = operationsString;
		operationsString ="";
		return aux;
	}
	
	/**
	 * Sets the reduce string. The leaves that compose the set <code>s</code>.
	 * @param s Set to be added to the tree.
	 */
	public void setReduceString(int[] s)
	{
		reduceString = "Reduced: { ";
		for (int i=0; i<s.length; i++)
		{
			reduceString += String.valueOf(s[i]) + " ";
		}
		reduceString += "}\n";
	}
	
	/**
	 * Verifies if a reduction is already concluded.
	 * @return <code>true</code>, if the reduction is
	 * concluded, or <code>false</code>, if it isn't.
	 */
	public boolean isDoneReduction() 
	{ 
		return doneReduction;
	}
	
	/**
	 * Signals the controller to paint the tree, and
	 * waits for signal to continue with the next step
	 * of the reduction.
	 */
	public void step()
	{
		if (stepByStep)
		{
			addSetLock.lock();
			controller.setToBePainted = true;
			canBePainted.signalAll();
			try
			{
				while (!painted)
				{
					nextStep.await();
				}
				painted = false;
				addSetLock.unlock();
			}
			catch(InterruptedException e)
			{
				//System.out.println("Error :" + e.toString());
			}
		}
	}
	
	/**
	 * Sets the lock and conditions to work with the controller.
	 * @param addSetLock A lock to works with some conditions.
	 * @param canAddSet Condition to adds a set to the tree.
	 * @param canBePainted Condition to paints the tree.
	 * @param nextStep Condition to executes the next step.
	 */
	public void setLocks(Lock addSetLock, Condition canAddSet, Condition canBePainted, Condition nextStep)
	{
		this.addSetLock = addSetLock;
		this.canAddSet = canAddSet;
		this.canBePainted = canBePainted;
		this.nextStep = nextStep;
	}
	
	/**
	 * Sets the tree to reduct completely or by steps.
	 * @param s <code>true</code> if reduction
	 * should be make step by step, drawing the tree,
	 * or <code>false</code> if should do the reduction
	 * completely.
	 */
	public void setStep(boolean s)
	{
		stepByStep = s;
	}
	
	/**
	 * Sets the tree to painted or not yet painted.
	 * @param p <code>true</code> if the tree was already
	 * painted, or <code>false</code> if it wasn't.
	 */
	public void setPainted(boolean p)
	{
		painted = p;
	}
	
	/**
	 * Resets the tree to a P-universal.
	 */
	public void reset()
	{
		doneReduction = true;
		root = new PQRNode(TYPE_PNODE, null);
		int i = 0;
		try
		{
			for ( i=0; i<leaves.length; i++ )
			{
				Leaf aLeaf = new Leaf(((Leaf)leaves[i]).getData(), root);
				root.addChild(aLeaf);
				leaves[i] = aLeaf;
			}
		}
		catch (OutOfMemoryError e)
		{
			System.out.println("Insufficient memory while creating leaf node #" + i);
			throw e;
		}
		hasChanged = true;
	}
}
