package pqrtreeapplet;

import javax.swing.*;
import java.awt.event.*;
import pqrtree.PQRTree;
import pqrtree.Node;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

/** 
 * The controller component of the model/view/controller for this application
 * This class starts the execution and either loads the GUI.
 * 
 * Based in the PQController class of Jon Harris' Graphical Implementation of 
 * PQ Tree Data Structure.
 * <a href="http://www.jharris.ca/portfolio/accode/PQTree.html">PQ Tree</a>
 * link.
 * 
 * @version 0.1
 * @author Bruno Dilly
 * 
 */
public class PQRController implements ActionListener
{
	private PQRTree pqrTree;
	private PQRWindow window;
	private Timer timer;
	public Lock addSetLock;
	public Condition canAddSet, canBePainted, nextStep;
	public boolean setToBePainted;
	private ExecutorService application;
	
	// this value specifies whether or not to draw the tree after
	// each and every template match.
	private boolean drawAllSteps = false;
	// this value specifies the delay in seconds after any redrawing
	// of the tree,
	private int delayInSeconds = 1;
	
	public PQRController(int universeSize) throws Exception
	{
		Object[] u = new Object[universeSize];
		for (int i = 0; i<universeSize; i++)
		{
			u[i]= new Integer(i);
		}
		addSetLock = new ReentrantLock();
		canAddSet = addSetLock.newCondition(); 
		canBePainted = addSetLock.newCondition(); 
		nextStep = addSetLock.newCondition();
		pqrTree = new PQRTree(u, this, addSetLock, canAddSet, canBePainted, nextStep);
		application = Executors.newFixedThreadPool(1);
		application.execute(pqrTree);
		init();
	}
	
	/**
	 * Creates a new window. Adds listeners to the buttons.
	 * @throws Exception
	 */
	public void init() throws Exception
	{
		window = new PQRWindow(pqrTree, drawAllSteps, delayInSeconds);
		window.getReduceButton().addActionListener(this);
		window.getStepButton().addActionListener(this);
		window.getResetButton().addActionListener(this);
		timer = new Timer(0, this);
	}
	
	/**
	 * Gets the window created by the controller.
	 * @return The window created.
	 */
	public PQRWindow getWindow() 
	{ 
		return window; 
	}
	
	/**
	 * Reduces the tree using the constraints provided by the GUI.
	 * @return An array of constraints. 
	 */
	private int[] getConstraintsFromGUI()
	{
		Object[] reductionArray = window.getReduceList().getSelectedValues();
		int[] constraints = new int[reductionArray.length];
		if ( reductionArray.length > 0 )
		{
			//boolean previouslyDoneReduction = pqrTree.isDoneReduction();      
			for (int i=0; i<reductionArray.length; i++)
			{
				constraints[i] = Integer.parseInt(reductionArray[i].toString()); 
			}
		}
		return constraints;
	}
	
	/**
	 *  Reduce the tree using the constraints provided in a vector.
	 * @param constraints The set to be added to the tree.
	 * @throws Exception
	 */
	public void reduceTree(int[] constraints) throws Exception
	{
		if (pqrTree.isDoneReduction())
		{
			setToBePainted = false;
			addSetLock.lock();
			pqrTree.setStep(false);
			pqrTree.addSet(constraints);
			canAddSet.signalAll();
			try
			{
				while(!setToBePainted)
				{
					canBePainted.await();
				}
				setToBePainted = false;
				window.refreshTree(pqrTree, drawAllSteps);
				window.getOutputArea().append(pqrTree.getOperationsString());
				window.getOutputArea().append(pqrTree.getReduceString());
				addSetLock.unlock();
			}
			catch(InterruptedException e)
			{
				//System.out.println("Error :" + e.toString());
			}
		}
		else
		{
			pqrTree.setStep(false);
			reduceTreeByStep();
		}
	}
	
	/**
	 * Reduces the tree by one step.
	 * @throws Exception
	 */
	public void reduceTreeByStep() throws Exception
	{
		setToBePainted = false;
		addSetLock.lock();
		if (pqrTree.isDoneReduction())
	    {
		  window.getReduceList().setEnabled(false);
		  pqrTree.setStep(true);
		  pqrTree.setPainted(false);
	      pqrTree.addSet(getConstraintsFromGUI());
	      canAddSet.signalAll();
	    }
		try
		{
			pqrTree.setPainted(true);
			nextStep.signalAll();
			while(!setToBePainted)
			{
				canBePainted.await();
			}
			setToBePainted = false;
			window.refreshTree(pqrTree, drawAllSteps);
			window.getOutputArea().append(pqrTree.getOperationsString());
			window.getOutputArea().append(pqrTree.getReduceString());
			if (pqrTree.isDoneReduction())
			{
				window.getOutputArea().append(pqrTree.getReduceString());
				window.getReduceList().setEnabled(true);
			}
			addSetLock.unlock();
		}
		catch(InterruptedException e)
		{
			//System.out.println("Error :" + e.toString());
		}
	}
		
	/**
	 * Resets the tree back to the universal tree.
	 * Concludes atual reduction, if necessary, for thread
	 * consistency.
	 * @throws Exception
	 */
	public void resetTree() throws Exception
	{	
		//kills threads
		//destroy();
		//pqrTree.reset();
		//restarts threads
		//application = Executors.newFixedThreadPool(1);
		//application.execute(pqrTree);
		
		//concludes reduction, if necessary
		if (!pqrTree.isDoneReduction())
		{
			reduceTree(getConstraintsFromGUI());
		}
		pqrTree.reset();
		window.getReduceList().setEnabled(true);
		window.getReduceButton().setEnabled(true);
		window.getStepButton().setEnabled(true);
		window.refreshTree(pqrTree, false);
		window.getOutputArea().setText("");
		timer.stop();
		//runs garbage collector
		System.gc();
	}
	
	/**
	 * Gets the leaf at a specific position of the
	 * leaves vector.
	 * @param index Position of the vector.
	 * @return Leaf at the position <code>index</code>.
	 */
	public Node getLeafAt(int index)
	{
		return pqrTree.getLeafAt(index);
	}
	
	/**
	 *  button click handlers for the gui.
	 *  @param e The action event occurred.
	 */
	public void actionPerformed(ActionEvent e)
	{
		try
		{
			if ( e.getSource() == window.getReduceButton() )
			{
		        timer.stop();
				if ( window.getPauseBox().isSelected() )
				{
					drawAllSteps = true;
					int pauseInSeconds = 0;
					try
					{
						pauseInSeconds = Integer.valueOf(window.getPauseField().getText()).intValue();
						if (pauseInSeconds < 1 )
							throw new NumberFormatException();
						delayInSeconds = pauseInSeconds;
					}
					catch (NumberFormatException nfe)
					{
						JOptionPane.showMessageDialog(window, "Please enter a valid postive numeric delay time");
						return;
					}
					
					if (pqrTree.isDoneReduction())
					{
						if (window.getReduceList().isSelectionEmpty())
						{
							JOptionPane.showMessageDialog(window, "Please select some elements for reduction");
							return;
						}
					}
					 timer.setDelay(delayInSeconds*1000);
			          timer.start();
				}
				else
				{
					if (window.getReduceList().isSelectionEmpty())
					{
						JOptionPane.showMessageDialog(window, "Please select some elements for reduction");
						return;
					}
					else
					{
						drawAllSteps = false;
						delayInSeconds = 0;
						reduceTree(getConstraintsFromGUI());
					}
				}
				
			}
			else if ( e.getSource() == window.getResetButton() )
			{
				resetTree();
			}
			 else if ( e.getSource() == window.getStepButton() )
		      {
		        if (window.getReduceList().isSelectionEmpty())
		        {
		          JOptionPane.showMessageDialog(window, "Please select some elements for reduction");
		          return;
		        }
		        timer.stop();
		        reduceTreeByStep();
		      }
		      else if ( e.getSource() == timer )
		      {
		        reduceTreeByStep();
		        if (pqrTree.isDoneReduction())
		        {
		          timer.stop();
		        }
		      }
		}
		catch(Exception ex)
		{
			ex.printStackTrace();
		}
	}  
	
	/**
	 * Kills the thread.
	 */
	public void destroy()
	{
		application.shutdown();
	}
}
